#include "fsprotect.h"

static const int MAX_OPEN_FILES = 4096;
static const int BLOCKSIZE = 4096;	//in bytes
static const int CSUMSIZE  = 16;	//in bytes
static const int DEBUG = 0;
static const int MHEAD = 0;
static void *handle = NULL;
static syscall_t *stable;
static meta_t **mtable;

static char* FSPROTECT_HOME;// = "/afs/cs.wisc.edu/u/f/i/fiedler/workspace/project4/meta";
//static char* FSPROTECT_HOME = "/home/fiedler/workspace/project4/meta";

/* to fix
 * 1) O_TRUNC
 * 2) O_APPEND
 * 3) 10000 files
 * 4) FS_HOMEWITHOUT '/'
 */


/*******************************META_T FUNCTIONS***********************************/
static int META_PATHSIZE;

meta_t *
meta_alloc(int fd, const char *path) {
	if(MHEAD){ printf("META_ALLOC\n"); fflush(stdout);}
	struct stat st;
	fstat(fd, &st);
	
	meta_t *m = (meta_t *)Malloc(sizeof(meta_t));
	
	m->fd = fd;
	m->size = st.st_size/BLOCKSIZE;
	m->cur_block = 0;
	
	m->path   = (char *)Malloc(strlen(path) + 1);
	m->c_path = (char *)Malloc(META_PATHSIZE);
	m->p_path = (char *)Malloc(META_PATHSIZE);
	
	strcpy(m->path, path);
	sprintf(m->c_path, "%s%llu.%lluc", FSPROTECT_HOME,
		 (unsigned long long int)st.st_dev, (unsigned long long int)st.st_ino);
	sprintf(m->p_path, "%s%llu.%llup", FSPROTECT_HOME,
		 (unsigned long long int)st.st_dev, (unsigned long long int)st.st_ino);
		 
	return m;
}

void
meta_free(meta_t *m) {
	if(MHEAD){ printf("META_FREE\n"); fflush(stdout);}
	free(m->path);
	free(m->c_path);
	free(m->p_path);
	free(m);
	//if(MHEAD){ printf("END META_FREE\n"); fflush(stdout);}
}


/***************************CONSTRUCTOR AND DESTRUCTOR*******************************/
void 
fs_init() {
	char *tmp;// = "/afs/cs.wisc.edu/u/f/i/fiedler/workspace/project4/meta";
	if((tmp = getenv("FSPROTECT_HOME")) == NULL){
		printf("error: must set FSPROTECT_HOME\n");
		exit(1);
	}
	
	
	FSPROTECT_HOME = (char *)Malloc((strlen(tmp) + 2) * sizeof(char));
	sprintf(FSPROTECT_HOME, "%s/", tmp);
		
	mtable = (meta_t **)Malloc(MAX_OPEN_FILES * sizeof(meta_t *));
	
	//size of string to hold meta pathnames
	META_PATHSIZE = sizeof(dev_t) + sizeof(ino_t) + strlen(FSPROTECT_HOME) + 10;
	//meta_pathname = Malloc(meta_pathsize * sizeof(char));
	
	// get handle to libc
	handle = dlopen("/lib/libc.so.6",  RTLD_LAZY);
	assert(handle != NULL);

	// set up syscall table
	stable				= (syscall_t *) Malloc(sizeof(syscall_t));
	stable->open		= (int(*)(const char *, int, ...)) dlsym(handle, "open");
	stable->close		= (int(*)(int)) dlsym(handle, "close");
	stable->read		= (ssize_t(*)(int, void *, size_t)) dlsym(handle, "read");
	stable->write		= (ssize_t(*)(int, const void *, size_t)) dlsym(handle, "write");
	stable->lseek		= (off_t(*)(int, off_t, int)) dlsym(handle, "lseek");
	stable->unlink		= (int(*)(const char *)) dlsym(handle, "unlink");
	stable->truncate	= (int(*)(const char *, off_t)) dlsym(handle, "truncate");
	stable->ftruncate	= (int(*)(int, off_t)) dlsym(handle, "ftruncate");
	
	//set up meta data table
	int i;
	for(i = 0; i < MAX_OPEN_FILES; i++) {
		mtable[i] = NULL;
	}
}

void 
fs_fini() {
	free(FSPROTECT_HOME);
	free(mtable);
	free(stable);
    dlclose(handle);
}

/*****************************FILE SYSTEM FUNTIONS***********************************/

/* make parity file from file parity */
void 
fs_make_parity(int fd) {
	if(MHEAD){ printf("FS_MAKE_PARITY\n"); fflush(stdout);}
	int i;
	char par[BLOCKSIZE];
	char buf[BLOCKSIZE];
	unsigned int *p = (unsigned int *) par;
    unsigned int *b = (unsigned int *) buf;
    
	for (int i = 0; i < 1024; i++) {
		*p = 0; p++;
		*b = 0; b++;
    }
    
    //reset pointers
	p = (unsigned int *) par;
	b = (unsigned int *) buf;
	
	// open file and parity data
	int tmp_fd   = Open(mtable[fd]->path, O_RDONLY);
	int meta_fdp = Open(mtable[fd]->p_path, O_CREAT | O_TRUNC | O_WRONLY, 0666);
	
	for(i = 0; i < mtable[fd]->size; i++) {
		if(DEBUG) {printf("ploop/n");}
		Read(tmp_fd, buf, BLOCKSIZE);
		//xor the bits
		for (int i = 0; i < 1024; i++) {
			*p ^= *b;
			p++; b++;
	    }
		p = (unsigned int *) par;
		b = (unsigned int *) buf;
	}
	
	Write(meta_fdp, par, BLOCKSIZE);
	Close(tmp_fd);
	Close(meta_fdp);
}

/* reconstruct single block w/ parity */
int 
fs_reconstruct(int fd, int bad_block) {
	if(MHEAD){ printf("FS_RECONSTRUCT bad_block = %d\n", bad_block); fflush(stdout);}
	int i;
	char fixed_block[BLOCKSIZE];
	char buf[BLOCKSIZE];
	unsigned int *f = (unsigned int *) fixed_block;
    unsigned int *b = (unsigned int *) buf;
    
	for (int i = 0; i < 1024; i++) {
		*f = 0; f++;
		*b = 0; b++;
    }
    
    // reset pointers
	f = (unsigned int *) fixed_block;
	b = (unsigned int *) buf;
	
	// open file and meta data
	int tmp_fd   = Open(mtable[fd]->path,   O_RDWR);
	int meta_fdc = Open(mtable[fd]->c_path, O_RDWR);
	int meta_fdp = Open(mtable[fd]->p_path, O_RDONLY);
	
	int rc;
	for(i = 0; i < mtable[fd]->size; i++) {
		//printf("loop %d", i);
		Read(tmp_fd, buf, BLOCKSIZE);
		if(i != bad_block) {	//skip the bad ones, just like santa
			if((rc = fs_check_sum(fd, i)) == 0) { //multiple bad blocks, welcome to shit creek
				//fprintf(stderr, "error: reconstruct: multiple bad bad blocks\n"); fflush(stderr);
				break;
			}
			//printf("use block");
			//xor the bits
			for (int i = 0; i < 1024; i++) {
				*f ^= *b;
				f++; b++;
		    }
			f = (unsigned int *) fixed_block;
			b = (unsigned int *) buf;
		}
		//printf("\n");
	}
	
	if(rc != 0) { //there is only one corrupt block FIX THAT SHIT
		// dont forget the parity
		//printf("add parity\n");
		Read(meta_fdp, buf, BLOCKSIZE);
		for (int i = 0; i < 1024; i++) {
			*f ^= *b;
			f++; b++;
	    }
		
		// fix the broken block. YAY!!!!
		Lseek(tmp_fd, bad_block * BLOCKSIZE, SEEK_SET);
		Write(tmp_fd, fixed_block, BLOCKSIZE);
	}
	
	Close(tmp_fd);
	Close(meta_fdc);
	Close(meta_fdp);
	return rc;
}

void 
fs_make_all_sum(int fd) {
	if(MHEAD){ printf("FS_MAKE_ALL_SUM\n"); fflush(stdout);}
	int i;
	md5_context md5con;
	unsigned char buf[BLOCKSIZE];
	unsigned char md5sum[16];
	
	// open file and meta data
	int tmp_fd   = Open(mtable[fd]->path, O_RDONLY);
	int meta_fdc = Open(mtable[fd]->c_path, O_CREAT | O_TRUNC | O_WRONLY, 0666);
	
	for(i = 0; i < mtable[fd]->size; i++) {
		//printf("loop\n");
		Read(tmp_fd, buf, BLOCKSIZE);
		md5_starts(&md5con);
		md5_update(&md5con, (uint8 *)buf, BLOCKSIZE);
		md5_finish(&md5con, md5sum);
		Write(meta_fdc, md5sum, CSUMSIZE);
	}
	Close(tmp_fd);
	Close(meta_fdc);
}

void 
fs_make_sum(int fd, int block_off) {
	if(MHEAD){ printf("FS_MAKE_SUM\n"); fflush(stdout);}
	md5_context md5con;
	unsigned char buf[BLOCKSIZE];
	unsigned char md5sum[16];
	
	// open file and meta data
	int tmp_fd   = Open(mtable[fd]->path, O_RDONLY);
	int meta_fdc = Open(mtable[fd]->c_path, O_WRONLY);
	
	Lseek(tmp_fd, block_off * BLOCKSIZE, SEEK_SET);
	Read(tmp_fd, buf, BLOCKSIZE);
	
	md5_starts(&md5con);
	md5_update(&md5con, (uint8 *)buf, BLOCKSIZE);
	md5_finish(&md5con, md5sum);
	
	Lseek(meta_fdc, block_off * CSUMSIZE, SEEK_SET);
	Write(meta_fdc, md5sum, CSUMSIZE);
	
	Close(tmp_fd);
	Close(meta_fdc);
}


int 
fs_check_all_sum(int fd) {
	if(MHEAD){ printf("FS_CHECK_ALL_SUM\n"); fflush(stdout);}
	int i;
	md5_context md5con;
	unsigned char old_sum[16];
	unsigned char new_sum[16];
	unsigned char buf[BLOCKSIZE];
	
	//open checksum data
	int tmp_fd   = Open(mtable[fd]->path, O_RDONLY);
	int meta_fdc = Open(mtable[fd]->c_path, O_RDONLY);
	
	for(i = 0; i < mtable[fd]->size; i++) {
		Read(meta_fdc, old_sum, CSUMSIZE);
		
		Read(tmp_fd, buf, BLOCKSIZE);
		md5_starts(&md5con);
		md5_update(&md5con, (uint8 *)buf, BLOCKSIZE);
		md5_finish(&md5con, new_sum);
		
		if(!(memcmp(old_sum, new_sum, CSUMSIZE) == 0)) {
			break;	//bad block
		}
	}
	
	if(i == mtable[fd]->size) {
		i = -1;
	}
	
	Close(tmp_fd);
	Close(meta_fdc);
	
	return i;
}


int 
fs_check_sum(int fd, int block_off) {
	if(MHEAD){ printf("FS_CHECK_SUM\n"); fflush(stdout);}
	md5_context md5con;
	unsigned char buf[BLOCKSIZE];
	unsigned char old_sum[16];
	unsigned char new_sum[16];
	
	//open checksum data
	int meta_fdc = Open(mtable[fd]->c_path, O_RDONLY);
	Lseek(meta_fdc, block_off * CSUMSIZE, SEEK_SET);
	Read(meta_fdc, old_sum, CSUMSIZE);
	
	// in case file is open in O_WRONLY open it up again for reading
	int tmp_fd = Open(mtable[fd]->path, O_RDONLY);
	Lseek(tmp_fd, block_off * BLOCKSIZE, SEEK_SET);
	Read(tmp_fd, buf, BLOCKSIZE);
	
	md5_starts(&md5con);
	md5_update(&md5con, (uint8 *)buf, BLOCKSIZE);
	md5_finish(&md5con, new_sum);
	
	Close(tmp_fd);
	Close(meta_fdc);
	
	return (memcmp(new_sum, old_sum, CSUMSIZE) == 0);
}

/*********************************SYSTEM FUNTIONS**************************************/

int 
open(const char *pathname, int flags, ...){
	if(MHEAD){ printf("OPEN\n"); fflush(stdout);}
	int fd;
	int exist = 0;
	
	if (flags & O_CREAT) { 
		// if O_CREAT is set, look for the mode argument too...
		va_list arg;
		va_start(arg, flags);
		mode_t mode = va_arg(arg, int);
		va_end(arg);
		//call the real open
		if(DEBUG) {printf("opening ext: %s\n", pathname); fflush(stdout);}
		fd = (*stable->open)(pathname, flags, mode);
	} else {
		//call the real open
		if(DEBUG) {printf("opening: %s\n", pathname); fflush(stdout);}
		fd = (*stable->open)(pathname, flags);
	}
	
	if(fd != -1)	//the file was opened real cool like
	{	//set up new meta data
		int rc;
		mtable[fd] = meta_alloc(fd, pathname);
		mtable[fd]->append = (flags & O_APPEND) ? 1 : 0;
		
		//check if it already exits
		if((rc = (*stable->open)(mtable[fd]->c_path, O_CREAT | O_EXCL | O_RDONLY, 0666)) == -1) {
			if(errno == EEXIST) { //set flag
				exist = 1;
				//if the file has been truncated to nothing erase old sums/parity
				if(flags & O_TRUNC)
				{	Unlink(mtable[fd]->c_path);
					Unlink(mtable[fd]->p_path);
				}
			} else { //read failed
				perror("error: open: failed to open meta file\n");
				exit(1);
			}
		} else { //bootstrap
			if(DEBUG) {printf("BOOT %d\n", rc); fflush(stdout);}
			Close(rc);
			//Open(mtable[fd]->c_path, O_CREAT, 0666);
			//Open(mtable[fd]->p_path, O_CREAT, 0666);
			//Close(mtable[fd]->c_path);
			//Close(mtable[fd]->p_path);
			fs_make_all_sum(fd);
			fs_make_parity(fd);
		}

		if(exist == 1) { //if meta info already exists
			int bad_block;
			
			//check file and try to reconstruct if reconstruct fails, whatever, we tried
			if((bad_block = fs_check_all_sum(fd)) != -1) {
				fs_reconstruct(fd, bad_block);
			}
		}
	}
	//if(MHEAD){ printf("END OPEN\n"); fflush(stdout);}
	return fd;
}

int 
close(int fd){
	if(MHEAD){ printf("CLOSE\n"); fflush(stdout);}
	//remove meta struct
	meta_free(mtable[fd]);
	mtable[fd] = NULL;
	 
	// do the close
	if(DEBUG) {printf("closing: %d\n", fd); fflush(stdout);}
	return (*stable->close)(fd);
}


off_t 
lseek(int fildes, off_t offset, int whence){
	int block_off = offset / BLOCKSIZE;
	//int old_block = mtable[fildes]->cur_block;
	
	int rc = (*stable->lseek)(fildes, offset, whence);
	
	if(rc != -1) { //if seek is sucessful
		switch(whence) {
			case SEEK_SET: //offset = offset
				mtable[fildes]->cur_block = block_off;
				break;
			case SEEK_CUR: //offset = offset + cur loc
				mtable[fildes]->cur_block += block_off;
				break;
			case SEEK_END: //offset = size of file + offset
				mtable[fildes]->size += block_off;
				break;
			default:
				//not a correct seek type
				break;
		}
		if(mtable[fildes]->cur_block > mtable[fildes]->size) {//lseek is outside of file
			//make an empty block
			unsigned char md5sum[16];
			md5_context md5con;
			char empty_block[BLOCKSIZE];
			unsigned int *e = (unsigned int *) empty_block;
    		
			for (int i = 0; i < 1024; i++) {
				*e = 0; e++;
		    }
		    
		    //make checksum for empty block
		    md5_starts(&md5con);
			md5_update(&md5con, (uint8 *)empty_block, BLOCKSIZE);
			md5_finish(&md5con, md5sum);
						
			// open meta data
			int meta_fdc   = Open(mtable[fildes]->c_path, O_WRONLY);
			
			//make new checksums
			for(int i = mtable[fildes]->size; i < mtable[fildes]->cur_block; i++) {
				Lseek(meta_fdc, i * CSUMSIZE, SEEK_SET);
				Write(meta_fdc, md5sum, CSUMSIZE);		
			}
			
			Close(meta_fdc);
		}
	}
	return rc;
}


ssize_t 
read(int fd, void *buf, size_t count) {
	int corrupt = 0;
	if(MHEAD){ printf("READ\n"); fflush(stdout);}
	//block out of range
	int i;
	int num_blocks = count / BLOCKSIZE;
	int end_block = num_blocks + mtable[fd]->cur_block;
	
	for(i = 0; i < num_blocks && mtable[fd]->cur_block < mtable[fd]->size; i++) {
		if(fs_check_sum(fd, mtable[fd]->cur_block) != 0) {
			mtable[fd]->cur_block++;
		} else {
			int rc;
			if((rc = fs_reconstruct(fd, mtable[fd]->cur_block)) != 0) {	//reconstruct
				mtable[fd]->cur_block = end_block;	//already know rest of blocks are ok
				break;								//exit loop
			} else {	//cant reconstruct
				corrupt = 1;
				//fprintf(stderr, "error: read: corrupt blocks unable to read from file\n");
			}
		}
	}
  	
	if(DEBUG) {printf("reading fd: %d\n", fd); fflush(stdout);}
	if(corrupt == 0) {
		return (*stable->read)(fd, buf, count);
	} else {
		errno = EIO;
		return -1;
	}
}


ssize_t 
write(int fd, const void *buf, size_t count){
	if(MHEAD){ printf("WRITE\n"); fflush(stdout);}
	if(DEBUG) {printf("writing fd: %d %s\n", fd, mtable[fd]->path); fflush(stdout);}
	int i;
	//int bad_block;
	int last_block = mtable[fd]->size - 1;
	int num_blocks = count / BLOCKSIZE;
	//unsigned char old_blocks[count];
	
	/* we are assuming no corruption prior to write	
	if((bad_block = fs_check_all_sum(fd)) != -1)
	{	fs_reconstruct(fd, bad_block);
	}
	*/
	//int tmp_fd = Open(mtable[fd]->path, O_RDWR);
	//save old blocks
	//pread(tmp_fd, old_blocks, count, mtable[fd]->cur_block * BLOCKSIZE);
	
	//write new blocks
	int rc = (*stable->write)(fd, buf, count);
	
	/*
	if(rc == -1) { // rewrite old data?
		pwrite(tmp_fd, buf, count, mtable[fd]->cur_block * BLOCKSIZE);
	}
	*/
	
	struct stat st;
	fstat(fd, &st); 
	mtable[fd]->size = (st.st_size / BLOCKSIZE);
	if(rc != -1) {//write was successful
		if(DEBUG) {printf("write sucess\n"); fflush(stdout);}
		if(mtable[fd]->append == 1) {
			mtable[fd]->cur_block = mtable[fd]->size - 1;
			for(i = last_block; i < num_blocks; i++) { //make checksums for new blocks
				fs_make_sum(fd, i);
			}
		} else {
			for(i = 0; i < num_blocks; i++) { //make checksums for new blocks
				fs_make_sum(fd, mtable[fd]->cur_block);
				(mtable[fd]->cur_block)++;
			}
		}
	}
	
	fs_make_parity(fd);
		
	return rc;
}


/*  unlink  deletes  a  name from the filesystem. If that name was the last
    link to a file and no processes have the file open the file is  deleted
    and the space it was using is made available for reuse.
*/

int 
unlink(const char *pathname){
	int rc;
	struct stat st;
	int fd = Open(pathname, O_RDONLY);
	
	meta_t *m = meta_alloc(fd, pathname);
	fstat(fd, &st);
	
	Close(fd);
	
	rc = (*stable->unlink)(pathname);
	
	if(rc != -1 && st.st_nlink == 1) { //this is truly the last of the hard links
		//remove metadata
		Unlink(m->c_path);
		Unlink(m->p_path);
	}
	
	meta_free(m);
	
	return rc;
}

/*	The  truncate  and  ftruncate functions cause the regular file named by
	path or referenced by fd to be truncated to a size of precisely  length
	bytes.
	If  the  file  previously  was larger than this size, the extra data is
	lost.  If the file previously was shorter,  it  is  extended,  and  the
	extended part reads as zero bytes.
	The file pointer is not changed.
*/

int
truncate(const char *path, off_t length){
	int rc;
	int fd = open(path, O_WRONLY);
	//mtable[fd] = meta_alloc(fd, path);
	
	rc = ftruncate(fd, length);
	close(fd);
	
	return rc;
}

int
ftruncate(int fd, off_t length){
	int rc;
	int bad_block;
	int new_size = length / BLOCKSIZE;
	
	if((bad_block = fs_check_all_sum(fd)) != -1) {
		fs_reconstruct(fd, bad_block);
	}
	
	rc = (*stable->ftruncate)(fd, length);
	
	struct stat st;
	fstat(fd, &st);
	mtable[fd]->size = st.st_size / BLOCKSIZE;
	
	if(rc != -1)
	{	Truncate(mtable[fd]->c_path, new_size * CSUMSIZE);
		fs_make_parity(fd);	
	}
	
	return rc;
}



/**********************SYSTEM WRAPPERS**********************************/

int 
Open(const char *pathname, int flags, ...) {
	int rc;	
	if (flags & O_CREAT) { 
		// if O_CREAT is set, look for the mode argument too...
		va_list arg;
		va_start(arg, flags);
		mode_t mode = va_arg(arg, int);
		va_end(arg);

		if(DEBUG) {printf("Opening ext: %s", pathname); fflush(stdout);}
		if((rc = (*stable->open)(pathname, flags, mode)) == -1)
		{	fprintf(stderr, "file %s: ", pathname); 
			perror(" error Opening file \n");
			exit(1);
		}
		//printf("%d\n", rc);
	} else {
		if(DEBUG) {printf("Opening: %s", pathname); fflush(stdout);}
		if((rc = (*stable->open)(pathname, flags)) == -1)
		{	fprintf(stderr, "file %s: ", pathname);
			perror(" error Opening file\n");
			exit(1);
		}
		//printf("%d\n", rc);
	}
	return rc;
}

int 
Close(int fd) {
	int rc;
	if(mtable[fd] != NULL) {
		if(DEBUG) {printf("Closing: %d %s\n", fd, mtable[fd]->path); fflush(stdout);}
	} else {
		if(DEBUG) {printf("Closing: %d\n", fd); fflush(stdout);}
	}
	if((rc = (*stable->close)(fd)) == -1) {
		perror("error: error closing file\n");
		exit(1);
	}
	return rc;
}

off_t
Lseek(int fildes, off_t offset, int whence) {
	int rc;
	if(DEBUG) {printf("Lseeking: %d\n", fildes); fflush(stdout);}
	if((rc = (*stable->lseek)(fildes, offset, whence)) == -1) {
		perror("error: error lseeking file\n");
		exit(1);
	}
	return rc;
}

ssize_t 
Read(int fd, void *buf, size_t count) {
	int rc;
	if(DEBUG) {printf("Reading: %d\n", fd); fflush(stdout);}
	if((rc = (*stable->read)(fd, buf, count)) == -1) {
		perror("error: error reading file\n");
		exit(1);
	}
	return rc;
}
ssize_t 
Write(int fd, const void *buf, size_t count) {
	int rc;
	if(DEBUG) {printf("Writing: %d\n", fd); fflush(stdout);}
	if((rc = (*stable->write)(fd, buf, count)) == -1) {
		perror("error: error writing file\n");
		exit(1);
	}
	return rc;
}

int
Unlink(const char *pathname) {
	int rc;
	if(DEBUG) {printf("Unlinking: %s\n", pathname); fflush(stdout);}
	if((rc = (*stable->unlink)(pathname)) == -1) {
		perror("error: error unlinking file\n");
		exit(1);
	}
	return rc;
}

int
Truncate(const char *path, off_t length) {
	int rc;
	if(DEBUG) {printf("Truncating: %s\n", path); fflush(stdout);}
	if((rc = (*stable->truncate)(path, length)) == -1) {
		perror("error: error Truncating file\n");
		exit(1);
	}
	return rc;
}

int
Ftruncate(int fd, off_t length) {
	int rc;
	if(DEBUG) {printf("Ftruncating: %d\n", fd); fflush(stdout);}
	if((rc = (*stable->ftruncate)(fd, length)) == -1) {
		perror("error: error Ftruncating file\n");
		exit(1);
	}
	return rc;
}

/* Malloc wrapper */
void *
Malloc(size_t size)
{	void *ptr;
	if((ptr = malloc(size)) != NULL)
	{	return ptr;
	}
	else
	{	fprintf(stderr, "malloc: memory request failed\n");
		exit(1);
	}	
}
