#ifndef __FSPROTECT__
#define __FSPROTECT__

#define _XOPEN_SOURCE 500

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <assert.h>

#include <sys/types.h>
#include <sys/stat.h>

#include <dlfcn.h>

#include "md5.h"

typedef struct __syscall {
	int		(*open)(const char *, int, ...);
	int		(*close)(int); 
	ssize_t	(*read)(int, void *, size_t);
	ssize_t	(*write)(int, const void *, size_t);
	off_t	(*lseek)(int, off_t, int);
	int		(*unlink)(const char *);
	int 	(*truncate)(const char *, off_t);
	int 	(*ftruncate)(int, off_t);
} syscall_t;

typedef struct __meta_t {
	int fd;
	int append;
	int size;		//size in 4KB blocks
	int cur_block;
	char *path;
	char *c_path;
	char *p_path;
} meta_t;

void fs_init() __attribute__((constructor));
void fs_fini() __attribute__((destructor));

meta_t * meta_alloc(int fd, const char *path);
void meta_free(meta_t *m);

void fs_make_parity(int fd);
int  fs_reconstruct(int fd, int bad_block);
void fs_make_all_sum(int fd);
void fs_make_sum(int fd, int block_off);
int  fs_check_all_sum(int fd);
int  fs_check_sum(int fd, int block_off);

int open(const char *pathname, int flags, ...);
int close(int fd);
off_t lseek(int fildes, off_t offset, int whence);
ssize_t read(int fd, void *buf, size_t count);
ssize_t write(int fd, const void *buf, size_t count);
int unlink(const char *pathname);
int truncate(const char *path, off_t length);
int ftruncate(int fd, off_t length);

/* WRAPPERS */
int Open(const char *pathname, int flags, ...);
int Close(int fd);
off_t Lseek(int fildes, off_t offset, int whence);
ssize_t Read(int fd, void *buf, size_t count);
ssize_t Write(int fd, const void *buf, size_t count);
int Unlink(const char *pathname);
int Truncate(const char *path, off_t length);
int Ftruncate(int fd, off_t length);
void *Malloc(size_t size);

#endif
